import click

from kks.util.click import GroupedGroup
from kks.util.common import config_directory
from kks.cmd.auth import auth
from kks.cmd.convert import convert
from kks.cmd.gen import gen
from kks.cmd.hide import hide, unhide
from kks.cmd.init import init
from kks.cmd.lint import lint
from kks.cmd.open import open_
from kks.cmd.run import run
from kks.cmd.status import status
from kks.cmd.submit import submit
from kks.cmd.sync import sync
from kks.cmd.test import test_
from kks.cmd.top import top
from kks.cmd.upgrade import upgrade, update


def cleanup():
    for filename in ['cookies.pickle', 'targets.yaml']:
        file = config_directory() / filename
        if file.exists():
            file.unlink()


@click.group(cls=GroupedGroup)
def cli():
    """KoKoS helper tool"""
    cleanup()


class Commands:
    workspace = ('Workspace-related', 1)
    ejudge = ('Ejudge interaction', 2)
    solution = ('Solution testing', 3)
    other = ('Other', 4)


cli.add_command(init, group=Commands.workspace)
cli.add_command(sync, group=Commands.workspace)
cli.add_command(hide, group=Commands.workspace)
cli.add_command(unhide, group=Commands.workspace)

cli.add_command(auth, group=Commands.ejudge)
cli.add_command(open_, group=Commands.ejudge)
cli.add_command(status, group=Commands.ejudge)
cli.add_command(top, group=Commands.ejudge)
cli.add_command(submit, group=Commands.ejudge)

cli.add_command(lint, group=Commands.solution)
cli.add_command(run, group=Commands.solution)
cli.add_command(gen, group=Commands.solution)
cli.add_command(test_, group=Commands.solution)

cli.add_command(convert, group=Commands.other)
cli.add_command(upgrade, group=Commands.other)
cli.add_command(update, group=Commands.other)



import click

from kks.ejudge import Standings, TaskInfo, StandingsRow, TaskScore
from kks.util.ejudge import load_auth_data

KKS_STAT_API = 'https://kks.darkkeks.me/api'
KKS_STAT_TIMEOUT = 3  # seconds


def send_standings(standings):
    import requests
    from requests import RequestException

    auth_data = load_auth_data()
    if not auth_data:
        return False

    data = {}

    data.update(auth_data_to_dict(auth_data))
    data.update(standings_to_dict(standings))

    try:
        response = requests.post(f"{KKS_STAT_API}/send", json=data, timeout=KKS_STAT_TIMEOUT)
        return response.ok
    except RequestException:
        return False


def get_global_standings(user):
    import requests
    from requests import RequestException

    parameters = {}

    auth_data = load_auth_data()
    if auth_data:
        parameters.update(auth_data_to_dict(auth_data))

    try:
        response = requests.get(f"{KKS_STAT_API}/get", params=parameters, timeout=KKS_STAT_TIMEOUT)
    except RequestException as e:
        click.secho(f'Failed to receive global standings: {e}', color='red', err=True)
        return None

    if not response.ok:
        click.secho(f'Server response was not successful: {response.text}', color='red', err=True)
        return None

    json_response = response.json()
    standings = standings_from_dict(json_response['standings'])
    standings.fix_is_self(user, auth_data.contest_id)
    return standings


def auth_data_to_dict(auth_data):
    return {
        'contest_id': auth_data.contest_id,
        'login': auth_data.login,
    }


def standings_to_dict(standings):
    return {
        'standings': {
            'tasks': [{
                'contest': task.contest,
                'name': task.name,
            } for task in standings.tasks],
            'rows': [{
                'place': row.place,
                'user': row.user,
                'tasks': [{
                    'score': task.score,
                    'status': task.status,
                } for task in row.tasks],
                'solved': row.solved,
                'score': row.score,
                'is_self': row.is_self,
            } for row in standings.rows],
        }
    }


def standings_from_dict(standings):
    tasks = standings['tasks']
    rows = standings['rows']
    for row in rows:
        for task, row_task in zip(tasks, row['tasks']):
            row_task['contest'] = task['contest']

    return Standings(
        tasks=[
            TaskInfo(task['name'], task['contest'])
            for task in tasks
        ],
        rows=[
            StandingsRow(
                row['place'],
                row['user'],
                [
                    TaskScore(task['contest'], task['score'], task['status'])
                    for task in row['tasks']
                ],
                row['solved'],
                row['score'],
                row['is_self'],
                row['contest_id']
            )
            for row in standings['rows']
        ],
    )



import os
import sys
from collections import namedtuple
from itertools import groupby

import click
from click._compat import isatty, strip_ansi
from tqdm import tqdm

from kks.ejudge import Status, ejudge_standings, ejudge_summary, get_problem_info, extract_contest_name, get_group_id, \
    get_contest_id
from kks.util.ejudge import EjudgeSession
from kks.util.stat import send_standings, get_global_standings
from kks.util.storage import Cache, Config

Problem = namedtuple('Problem', ['href', 'short_name', 'contest'])  # used for caching the problem list

CONTEST_DELIMITER = ' | '


@click.command(short_help='Parse and display user standings')
@click.option('-l', '--last', type=int,
              help='Print result of last N contest')
@click.option('-a', '--all', 'all_', is_flag=True,
              help='Print result of all contests')
@click.option('-c', '--contest', 'contests', type=str, multiple=True,
              help='Print the results of the selected contest')
@click.option('-m', '--max', 'max_', is_flag=True,
              help='Print maximal possible scores (based on current deadlines)')
@click.option('-nc', '--no-cache', is_flag=True,
              help='Clear cache and reload task info (used with --max)')
@click.option('-g', '--global', 'global_', is_flag=True,
              help='Use global standings instead of group one. May be outdated')
@click.option('-f', '--group', 'groups', type=int, multiple=True,
              help='Print standings of selected groups')
@click.option('--global-opt-out', is_flag=True,
              help='Opt out from submitting your group results')
def top(last, all_, contests, groups, max_, no_cache, global_, global_opt_out):
    """
    Parse and display user standings

    \b
    Example usage:
        kks top
        kks top --all
        kks top -c sm01 -c sm02
        kks top --last 2
    """

    config = Config()

    if global_opt_out:
        if click.confirm(click.style('Do you really want to opt out from sending your group standings to '
                                     'kks.darkkeks.me?', bold=True, fg='red')):
            opt_out(config)
        return
    elif config.options.global_opt_out is None:
        init_opt_out(config)

    session = EjudgeSession()
    standings = ejudge_standings(session)

    if not config.options.global_opt_out:
        if not send_standings(standings):
            click.secho('Failed to send standings to kks api', fg='yellow', err=True)

    if global_:
        standings = get_global_standings(standings.user)
        if standings is None:
            click.secho('Standings are not available now :(', fg='yellow', err=True)
            return

        if groups:
            standings = filter_groups(standings, groups)
            if standings is None:
                return

    if max_:
        standings = estimate_max(standings, session, config, no_cache)

    display_standings(standings, last, contests, all_, global_)


def init_opt_out(config):
    click.secho('Standings can be sent for aggregation to kks api. '
                'This allows us to create global standings (you can try it out with kks top --global)', err=True)
    click.secho('You can always disable sending using kks top --global-opt-out', err=True)
    if click.confirm(click.style('Do you want to send group standings to kks api?', bold=True, fg='red'), default=True):
        click.secho('Thanks a lot for you contribution! We appreciate it!', fg='green', bold=True, err=True)
        config.options.global_opt_out = False
        config.save()
    else:
        opt_out(config)


def opt_out(config):
    click.secho('Successfully disabled standings sending. You can always enable sending by manually editing '
                '~/.kks/config.ini', color='red', err=True)
    config.options.global_opt_out = True
    config.save()


def display_standings(standings, last, contests, all_, global_):
    table = FancyTable()

    table.add_column(StaticColumn('Place', 6, lambda row: row.place))
    table.add_column(StaticColumn.padding(1))
    table.add_column(StaticColumn('User', 24, lambda row: row.user, right_just=False))
    table.add_column(StaticColumn('Solved', 6, lambda row: row.solved))
    table.add_column(StaticColumn('Score', 6, lambda row: row.score))

    if global_:
        table.add_column(StaticColumn('Group', 6, lambda row: get_group_id(row.contest_id)))

    terminal_width = get_terminal_width()

    if isatty(sys.stdout):
        contests_width = terminal_width - table.calc_width() - 1
        default_contest_count = \
            get_default_contest_count(standings.contests, standings.tasks_by_contest, contests_width)
    else:
        default_contest_count = len(standings.contests)

    contests = select_contests(standings, last, contests, all_, default_contest_count)
    if contests is None:
        return

    table.add_column(TasksColumn(contests, standings.tasks_by_contest))

    exceeds_width = table.calc_width() > terminal_width

    lines = table.render(standings.rows)
    output = '\n'.join(lines)

    if isatty(sys.stdout) and (exceeds_width or global_):
        if 'LESS' not in os.environ:
            os.environ['LESS'] = '-S -R'
        click.echo_via_pager(output)
    else:
        click.secho(output)


class Column:
    def header(self):
        raise NotImplemented()

    def value(self, row):
        raise NotImplemented()

    def width(self):
        raise NotImplemented()


class StaticColumn(Column):
    def __init__(self, name, width, mapper, right_just=True):
        self.name = name
        if self.name is None:
            self.name = ''
        self.mapper = mapper
        self.actual_width = max(width, len(self.name))
        self.right_just = right_just

    def _justify(self, value):
        return value.rjust(self.actual_width, ' ') if self.right_just \
            else value.ljust(self.actual_width, ' ')

    def header(self):
        return click.style(self._justify(self.name), fg='white', bold=True)

    def value(self, row):
        return click.style(self._justify(str(self.mapper(row))), fg=row.color(), bold=row.bold())

    def width(self):
        return self.actual_width

    @classmethod
    def padding(cls, width):
        # width - 1, так как лишняя колонка уже добавляет один пробел
        return cls(None, width - 1, lambda _: '')


class TasksColumn(Column):
    DELIMITER = ' | '

    def __init__(self, contests, tasks_by_contests):
        self.contests = contests
        self.contest_widths = get_contest_widths(self.contests, tasks_by_contests)

    def header(self):
        return ''.join([
            click.style(TasksColumn.DELIMITER, fg='white', bold=False) +
            click.style(contest.ljust(self.contest_widths[contest], ' '), fg='white', bold=True)
            for contest in self.contests
        ])

    def value(self, row):
        return ''.join([
            click.style(TasksColumn.DELIMITER, fg=row.color(), bold=row.bold()) + ' '.join([
                click.style('{:>3}'.format(task.table_score() or ''), fg=task.color(), bold=task.bold())
                for task in tasks
            ])
            for contest, tasks in groupby(row.tasks, lambda task: task.contest)
            if contest in self.contests
        ])

    def width(self):
        return sum(self.contest_widths.values()) + len(TasksColumn.DELIMITER) * len(self.contests)


class FancyTable:
    def __init__(self):
        self.columns = []

    def add_column(self, column):
        self.columns.append(column)

    def calc_width(self):
        content = sum([column.width() for column in self.columns])
        return content + len(self.columns) - 1

    def render(self, rows):
        lines = [
            ' '.join([
                column.header()
                for column in self.columns
            ])
        ]

        for row in rows:
            lines.append(' '.join([
                column.value(row)
                for column in self.columns
            ]))

        return lines


def select_contests(standings, last, contests, all_, default_count):
    has_last = last is not None
    has_contests = len(contests) > 0

    if sum([has_last, has_contests, all_]) > 1:
        click.secho('Arguments are exclusive, specify no more than one')
        return None

    if all_:
        return standings.contests

    if has_contests:
        def is_available(contest):
            if contest not in standings.contests:
                click.echo(
                    click.style('Contest ', fg='red') +
                    click.style(contest, fg='blue', bold=True) +
                    click.style(' not found!', fg='red'),
                    err=True
                )
            return contest in standings.contests

        return list(filter(is_available, contests))

    last = last or default_count

    return standings.contests[-last:] if last > 0 else []


def get_default_contest_count(contests, tasks_by_contest, max_width):
    """По ширине пытается определить, сколько колонок можно вывести"""

    delimiter_width = len(CONTEST_DELIMITER)
    contest_widths = get_contest_widths(contests, tasks_by_contest)

    width_sum = 0
    for i, contest in enumerate(contests[::-1]):
        width_sum += delimiter_width + contest_widths[contest]
        if width_sum > max_width:
            return i

    return len(contests)


def get_contest_widths(contests, tasks_by_contest):
    return {
        contest: len('100') * len(tasks_by_contest[contest]) + len(tasks_by_contest[contest]) - 1
        for contest in contests
    }


def get_terminal_width():
    (width, _) = click.get_terminal_size()
    return width


def filter_groups(standings, groups):
    group_contest_ids = []
    for group in groups:
        contest_id = get_contest_id(group)
        if contest_id is None:
            click.secho(f'Invalid group name: "{group}"', fg='red', err=True)
            return None
        group_contest_ids.append(contest_id)

    standings.rows = [
        row
        for row in standings.rows
        if row.contest_id in group_contest_ids
    ]

    for i, row in enumerate(standings.rows):
        row.place = i + 1

    return standings


def estimate_max(standings, session, config, force_reload):
    # NOTE may produce incorrect results for "krxx" contests (they may be reopened?)

    def cached_problem(problem):
        return Problem(problem.href, problem.short_name, extract_contest_name(problem.short_name))

    with Cache('problem_info', compress=True).load() as cache:
        if force_reload:
            cache.clear()

        problem_list = cache.get('problem_links', [])  # we can avoid loading summary

        with tqdm(total=len(standings.tasks), leave=False) as pbar:
            def with_progress(func, *args, **kwargs):
                result = func(*args, **kwargs)
                pbar.update(1)
                return result

            if len(problem_list) != len(standings.tasks) or \
                    any(problem.short_name != task.name for problem, task in zip(problem_list, standings.tasks)):
                problem_list = ejudge_summary(session)
                problem_list = [cached_problem(p) for p in problem_list]
                cache.set('problem_links', problem_list)
            problems = [with_progress(get_problem_info, problem, cache, session) for problem in problem_list]

    for row in standings.rows:
        for task_score, problem, orig_problem in zip(row.tasks, problems, problem_list):
            if task_score.score is None or task_score.score == '0':
                if task_score.status == Status.REJECTED:
                    max_score = problem.full_score
                else:
                    max_score = problem.full_score - problem.current_penalty
                    if task_score.score == '0':
                        max_score -= problem.run_penalty
                        # actually may be lower
                if config.options.max_kr and  orig_problem.short_name.startswith('kr') and task_score.status == Status.TESTING:
                    max_score = 200  # not always true, but we cannot get real max_score
                    if task_score.score == '0':  # at least one partial solution (are CE counted?)
                        max_score -= 10
                if max_score > 0:
                    row.solved += 1
                    row.score += max_score
                    task_score.score = max_score
                    task_score.status = Status.REVIEW

    standings.rows.sort(key=lambda x: (x.score, x.solved), reverse=True)
    for i, row in enumerate(standings.rows):
        row.place = i + 1

    return standings



import gzip
import pickle
import sys
from configparser import ConfigParser
from datetime import datetime, timedelta
from os import environ
from time import time

import click

from kks.util.common import Singleton, config_directory


class Section:
    """
    in config.ini all option names are lowercase and all underscores are replaced with dashes
    """
    def __init__(self, config, name):
        super().__setattr__('_config', config)
        super().__setattr__('_name', Section.canonical_name(name))

    @staticmethod
    def canonical_name(name):
        return name.capitalize()

    @staticmethod
    def to_option(name):
        return name.lower().replace('_', '-')

    def _convert(self, key, value):
        type_ = self.__annotations__[key]
        if type_ is bool:
            return self._config._convert_to_boolean(value)
        return type_(value)

    def _is_option(self, key):
        return key in super().__getattribute__('__annotations__')

    def _check_key(self, key):
        if not self._is_option(key):
            raise AttributeError(f'Option "{key}" is not allowed in config section "{self._name}"')

    def __getattribute__(self, key):
        if not super().__getattribute__('_is_option')(key):
            return super().__getattribute__(key)

        value = self._config.get(self._name, Section.to_option(key), fallback=None)
        if value is not None:
            return self._convert(key, value)
        return getattr(type(self), key, None)  # default

    def __setattr__(self, key, value):
        self._check_key(key)
        if not self._config.has_section(self._name):
            self._config.add_section(self._name)
        self._config.set(self._name, Section.to_option(key), str(value))

    def __delattr__(self, key):
        self._check_key(key)
        if not self._config.has_section(self._name):
            return
        self._config.remove_option(self._name, Section.to_option(key))


class EnvSection(Section):
    @staticmethod
    def to_envvar(name):
        return name.upper()

    def __getattribute__(self, key):
        if not super().__getattribute__('_is_option')(key):
            return super().__getattribute__(key)  # how to avoid second check?

        envvar = environ.get(EnvSection.to_envvar(key))
        if envvar is not None:
            return self._convert(key, envvar)
        return super().__getattribute__(key)


class AuthSection(Section):
    login: str
    password: str
    contest: int


class OptionsSection(EnvSection):
    save_html_statements: bool = True
    save_md_statements: bool = True
    mdwidth: int = 100
    max_kr: bool = False
    global_opt_out: bool


class ConfigModel:
    auth: AuthSection
    options: OptionsSection


class Config(metaclass=Singleton):
    """global kks config"""

    def __init__(self):
        self._file = config_directory() / 'config.ini'
        self._config = ConfigParser()
        if self._file.is_file():
            self._config.read(self._file)
        # delete legacy section
        if self._config.has_section('Links'):
            self._config.remove_section('Links')
            self.save()

    def save(self):
        with self._file.open('w') as f:
            self._config.write(f)

    def reload(self):
        """force reload from disk"""
        if self._file.is_file():
            self._config.read(self._file)
        else:
            self._config.clear()

    def __getattribute__(self, key):
        if key in ConfigModel.__annotations__:
            section_type = ConfigModel.__annotations__[key]
            return section_type(self._config, key)
        return super().__getattribute__(key)

    def __delattr__(self, key):
        if key in ConfigModel.__annotations__:
            self._config.remove_section(Section.canonical_name(key))
        else:
            super().__delattr__(key)


class PickleStorage:
    _service_keys = ('__version__',)

    def __init__(self, name, compress=False, version=1):
        self.name = name
        self.compress = compress

        suffix = '.pickle.gz' if compress else '.pickle'
        self._file = (config_directory() / name).with_suffix(suffix)
        self._version = version
        self._init_data()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.save()

    def _init_data(self, clean=True):
        self._data = {'__version__': self._version}
        self._clean = clean

    def load(self):
        self._init_data()

        if self._file.exists():
            try:
                with self._file.open('rb') as f:
                    data = f.read()
                if self.compress:
                    data = gzip.decompress(data)
                self._data.update(pickle.loads(data))
            except Exception:
                click.secho(f'Storage file {self._file} is corrupted', bg='red', err=True)
                if click.confirm(click.style('Erase all saved data?', fg='red', bold=True)):
                    self._init_data(False)
                else:
                    click.secho(f'You need to fix or delete {self._file.absolute()} manually', fg='red', err=True)
                    sys.exit()

        if self._data['__version__'] != self._version:
            click.secho(f'{self._file} uses an incompatible storage version, clearing saved data', bg='red', err=True)
            self._init_data(False)
        return self

    def keys(self):
        for k in self._data.keys():
            if k not in self._service_keys:
                yield k

    def get(self, key, default=None):
        return self._data.get(key, default)

    def set(self, key, value):
        self._data[key] = value
        self._clean = False

    def erase(self, key):
        # If key is not found, no errors are raised
        if key in self._data:
            del self._data[key]
            self._clean = False

    def clear(self):
        if len(self._data) > len(self._service_keys):
            self._init_data(False)

    def save(self):
        if self._clean:
            return
        with self._file.open('wb') as f:
            data = pickle.dumps(self._data)
            if self.compress:
                data = gzip.compress(data)
            f.write(data)


class Cache(PickleStorage):
    """Storage for data with expiration times"""
    def get(self, key, default=None):
        # default is returned if value is not found or outdated
        value, exp_time = super().get(key, (None, -1))
        if exp_time == -1:
            return default
        if exp_time is not None and time() > exp_time:
            self.erase(key)
            return default
        return value

    def set(self, key, value, expiration=None):
        """
        expiration may be None, datetime or timedelta
        If expiration is None, value will never become outdated
        """
        if expiration is None:
            exp_time = None
        elif isinstance(expiration, datetime):
            exp_time = int(expiration.timestamp())
        elif isinstance(expiration, timedelta):
            exp_time = int(time() + expiration.total_seconds())
        else:
            raise TypeError('Invalid argument type')

        super().set(key, (value, exp_time))
        self._clean = False
